#include <stdio.h>

  /* Prikaz realnih brojeva tipa double.
     Pretpostavka = "little endian" reprezentacija,
     tj. manje znacajni bitovi su na NIZIM adresama. */

  /* Broj se ucitava. */

  /* Napomena:
       Ovisno o opcijama kod prevodjenja, moze se dogoditi da funkcija printf
     NE zna korektno napisati broj. Primjer = 1.0e-310.
       Razlog: Opcija /O3 za optimizaciju ukljucuje i pretvaranje
     denormaliziranih brojeva u nulu ("flush denormal results to zero"). */


/****************************************************************************/


    /* String za ime realnog tipa. */
const char real_type_name[] = "double";


    /* Duljina tipa u byte-ovima. Broj bitova = length * 8. */
const unsigned int length = 8;

    /* Preciznost cijelog signifikanda (broj bitova = znamenki u bazi 2),
       ukljucivo i "skriveni" bit (hidden bit), ako postoji.
       Broj bitova u zavrsnom ("trailing", razlomljenom) dijelu
       signifikanda je t = prec_2 - 1, tj., najnizi bit signifikanda
       dolazi uz potenciju 2^(-t). */
const unsigned int prec_2 = 53;

    /* Potrebna preciznost u bazi 10 za korektan tzv.
       "write-read" ciklus, tako da dobijemo isti prikaz
       broja (u bazi 2), nakon ispisa i citanja (u bazi 10). */
const unsigned int prec_10 = 17;

    /* Logicki indikator = je li vodeci bit signifikanda "skriven"
       ili ne (tj. pamti se). Stvarni broj bitova za pamcenje
       signifikanda je q = prec_2 - hidden_bit. */
const int hidden_bit = 1;


    /* Broj bitova karakteristike, odnosno, binarnog eksponenta.
       Prema pretpostavci (little endian), ti bitovi se nalaze
       na najvisim adresama (odmah iza predznaka). */
const unsigned int exp_bits = 11;

    /* Maksimalna karakteristika = 2^(exp_bits) - 1. */
const unsigned int max_charact = 2047;

    /* Pomak za binarni eksponent = 2^(exp_bits - 1) - 1
       = max_charact / 2 (cjelobrojno dijeljenje). */
const int bias = 1023;


    /* Indeks bita predznaka (najvisi bit u cijelom prikazu)
       = length * 8 - 1, s tim da indekse brojimo od 0 (najnizi bit). */
const int sign_bit_ind = 63;

    /* Indeks vodeceg (najviseg) bita karakteristike (eksponenta)
       = sign_bit_ind - 1. */
const int exp_high_ind = 62;

    /* Indeks zadnjeg (najnizeg) bita karakteristike (eksponenta)
       = sign_bit_ind - exp_bits. */
const int exp_low_ind = 52;


/****************************************************************************/

    /* Tip za prikaz broja, tako da se lako dodje do bitova. */

typedef union {
    double number;            /* double = 64 bita. */
    unsigned char byte[8];    /* Polje od 8 byteova. */
} Number_Repr;

    /* Tip za vrijednosti pojedinih dijelova u prikazu broja,
       gledano kao triplet = (predznak, eksponent, signifikand).
       Uz stvarni binarni eksponent, pamtimo i karakteristiku. */

typedef struct {
    int sign;                /* Predznak = 0 ili 1. */
    unsigned int charact;    /* Karakteristika = pomaknuti eksponent. */
    int bin_exp;             /* Pravi eksponent baze 2 u prikazu. */
    double signif;           /* Stvarni signifikand. */
} Number_Triplet;


/****************************************************************************/

    /* Pretvaranje prikaza broja u triplet. */

Number_Triplet Repr_to_Triplet(Number_Repr repr)
{
    Number_Triplet triplet;
    int bit, i, byte_index, byte_pos;
    unsigned char byte_mask;
    unsigned int charact = 0u;
    double signif = 0.0;

        /* Predznak u prikazu = vodeci bit. */
    byte_index = length - 1;
    byte_pos = 7;
    byte_mask = 0x80u;
    triplet.sign = repr.byte[byte_index] & byte_mask ? 1 : 0;

        /* Pomak bit-pozicije za jedno mjesto nanize. */
    --byte_pos;
    byte_mask = byte_mask >> 1;

        /* Slaganje karakteristike po bitovima - od najviseg,
           do najnizeg. Koristimo Hornerov algoritam u bazi 2. */
    for (i = exp_high_ind; i >= exp_low_ind; --i) {

        bit = repr.byte[byte_index] & byte_mask ? 1 : 0;

            /* Pomak karakteristike za bit navise i dodatak novog bita. */
        charact = (charact << 1) + bit;

            /* Pomak bit-pozicije za jedno mjesto nanize. */
        if (byte_pos > 0) {
            --byte_pos;
            byte_mask = byte_mask >> 1;
        }
        else {
            --byte_index;
            byte_pos = 7;
            byte_mask = 0x80u;
        }
    }

    triplet.charact = charact;

        /* Binarni eksponent = karakteritika - pomak (bias). */
    triplet.bin_exp = triplet.charact - bias;

        /* Slaganje signifikanda po bitovima - od najnizeg,
           do najviseg (koji moze biti skriven). */
    byte_index = 0;
    byte_pos = 0;
    byte_mask = 0x01u;
    for (i = 0; i < exp_low_ind; ++i) {

            /* Pomak signifikanda za bit nanize = podijeli ga s 2. */
        signif = 0.5 * signif;

            /* Provjera bita s indeksom i. Ako je jednak 1,
               onda dodaj 1 signifikandu. */
        if ((repr.byte[byte_index] & byte_mask) > 0)
            signif = signif + 1.0;

            /* Pomak bit-pozicije za jedno mjesto navise. */
        if (byte_pos < 7) {
            ++byte_pos;
            byte_mask = byte_mask << 1;
        }
        else {
            ++byte_index;
            byte_pos = 0;
            byte_mask = 0x01u;
        }
    }

        /* Ako je vodeci (cjelobrojni) bit signifikanda skriven,
           onda ga otkrivamo iz karakteritike/eksponenta. */
    if (hidden_bit) {
            /* Pomak signifikanda za bit nanize = podijeli ga s 2. */
        signif = 0.5 * signif;
            /* Provjera skrivenog bita iz karakteristike. Ako je
               karakteristika pozitivna, skriveni bit je jednak 1,
               onda dodaj 1 signifikandu. */
        if (triplet.charact > 0)
            signif = signif + 1.0;
    }

    triplet.signif = signif;

    return triplet;
}

/****************************************************************************/

    /* Funkcija za ispis bitova u prikazu (sva tri dijela). */

/*
  Double:
    0 01111111111   1111 10000000 00000000 00000000 00000000 00000000 00000000
*/

void ispis_bitova(Number_Repr repr)
{
    int bit, i, byte_index, byte_pos;
    unsigned char byte_mask;

        /* Ispis bitova - od najviseg do najnizeg.
           Indeks najviseg bita u prikazu = ukupan broj bitova - 1. */
    byte_index = length - 1;
    byte_pos = 7;
    byte_mask = 0x80u;

        /* Vodeci bit = predznak. */
    bit = repr.byte[byte_index] & byte_mask ? 1 : 0;
    printf("    %1d ", bit);

        /* Pomak bit-pozicije za jedno mjesto nanize. */
    --byte_pos;
    byte_mask = byte_mask >> 1;

        /* Bitovi karakteristike. */
    for (i = exp_high_ind; i >= exp_low_ind; --i) {

        bit = repr.byte[byte_index] & byte_mask ? 1 : 0;
        printf("%1d", bit);

            /* Pomak bit-pozicije za jedno mjesto nanize. */
        if (byte_pos > 0) {
            --byte_pos;
            byte_mask = byte_mask >> 1;
        }
        else {
            --byte_index;
            byte_pos = 7;
            byte_mask = 0x80u;
        }
    }

    printf("   ");

        /* Bitovi signifikanda (bez skrivenog, ako ga ima). */
    for (i = exp_low_ind - 1; i >= 0; --i) {

        bit = repr.byte[byte_index] & byte_mask ? 1 : 0;
        printf("%1d", bit);

        if (i % 8 == 0 && i > 0) printf(" ");

            /* Pomak bit-pozicije za jedno mjesto nanize. */
        if (byte_pos > 0) {
            --byte_pos;
            byte_mask = byte_mask >> 1;
        }
        else {
            --byte_index;
            byte_pos = 7;
            byte_mask = 0x80u;
        }
    }

    printf("\n");

    return;
}

/****************************************************************************/

void ispis_prikaza(Number_Repr repr)
{
    Number_Triplet triplet;

        /* Pretvori zadani prikaz u triplet, za provjeru Inf/Nan. */
    triplet = Repr_to_Triplet(repr);


    printf("\n");

        /* Ispis kao broj. */
    printf("   Broj (%s) = ", real_type_name);

        /* Provjera za Inf/NaN. */
    if (triplet.charact == max_charact) {
        if (triplet.signif == 1.0) {    /* Inf. */
            if (triplet.sign == 0)
                printf("+Inf");
            else
                printf("-Inf");
        }
        else
            printf("NaN");
    }
    else {
            /* Korektni broj, ali moze biti triplet.charact = 0,
               tj. "gradual underflow". */
        printf("%*.*e",
               (int) (prec_10 + 7), (int) (prec_10 - 1), repr.number);
    }

    printf("\n\n");

        /* Ispis kao triplet. */
    printf("   Triplet = ( %1d, %5d [%4u], %*.*e )\n",
           triplet.sign, triplet.bin_exp, triplet.charact,
           (int) (prec_10 + 7), (int) (prec_10 - 1), triplet.signif);

        /* Ispis svih bitova u prikazu. */
    ispis_bitova(repr);

    return;
}

/****************************************************************************/

int main(void)
{
    Number_Repr repr;

    printf(" Upisi realni broj: ");
    scanf("%lf", &repr.number);

    ispis_prikaza(repr);

    return 0;
}
