#include <stdio.h>
#include <math.h>

  /* Prikaz realnih brojeva tipa double.
     Pretpostavka = "little endian" reprezentacija,
     tj. manje znacajni bitovi su na NIZIM adresama. */

  /* Primjeri prikaza brojeva i ostalih vrijednosti (Inf, Nan).
     Broj se moze zadati izravno i kao triplet. */

  /* Napomena:
       Ovisno o opcijama kod prevodjenja, moze se dogoditi da funkcija printf
     NE zna korektno napisati broj. Primjer = 1.0e-310.
       Razlog: Opcija /O3 za optimizaciju ukljucuje i pretvaranje
     denormaliziranih brojeva u nulu ("flush denormal results to zero"). */


/****************************************************************************/


    /* String za ime realnog tipa. */
const char real_type_name[] = "double";


    /* Duljina tipa u byte-ovima. Broj bitova = length * 8. */
const unsigned int length = 8;

    /* Preciznost cijelog signifikanda (broj bitova = znamenki u bazi 2),
       ukljucivo i "skriveni" bit (hidden bit), ako postoji.
       Broj bitova u zavrsnom ("trailing", razlomljenom) dijelu
       signifikanda je t = prec_2 - 1, tj., najnizi bit signifikanda
       dolazi uz potenciju 2^(-t). */
const unsigned int prec_2 = 53;

    /* Potrebna preciznost u bazi 10 za korektan tzv.
       "write-read" ciklus, tako da dobijemo isti prikaz
       broja (u bazi 2), nakon ispisa i citanja (u bazi 10). */
const unsigned int prec_10 = 17;

    /* Logicki indikator = je li vodeci bit signifikanda "skriven"
       ili ne (tj. pamti se). Stvarni broj bitova za pamcenje
       signifikanda je q = prec_2 - hidden_bit. */
const int hidden_bit = 1;


    /* Broj bitova karakteristike, odnosno, binarnog eksponenta.
       Prema pretpostavci (little endian), ti bitovi se nalaze
       na najvisim adresama (odmah iza predznaka). */
const unsigned int exp_bits = 11;

    /* Maksimalna karakteristika = 2^(exp_bits) - 1. */
const unsigned int max_charact = 2047;

    /* Pomak za binarni eksponent = 2^(exp_bits - 1) - 1
       = max_charact / 2 (cjelobrojno dijeljenje). */
const int bias = 1023;


    /* Indeks bita predznaka (najvisi bit u cijelom prikazu)
       = length * 8 - 1, s tim da indekse brojimo od 0 (najnizi bit). */
const int sign_bit_ind = 63;

    /* Indeks vodeceg (najviseg) bita karakteristike (eksponenta)
       = sign_bit_ind - 1. */
const int exp_high_ind = 62;

    /* Indeks zadnjeg (najnizeg) bita karakteristike (eksponenta)
       = sign_bit_ind - exp_bits. */
const int exp_low_ind = 52;


/****************************************************************************/

    /* Tip za prikaz broja, tako da se lako dodje do bitova. */

typedef union {
    double number;            /* double = 64 bita. */
    unsigned char byte[8];    /* Polje od 8 byteova. */
} Number_Repr;

    /* Tip za vrijednosti pojedinih dijelova u prikazu broja,
       gledano kao triplet = (predznak, eksponent, signifikand).
       Uz stvarni binarni eksponent, pamtimo i karakteristiku. */

typedef struct {
    int sign;                /* Predznak = 0 ili 1. */
    unsigned int charact;    /* Karakteristika = pomaknuti eksponent. */
    int bin_exp;             /* Pravi eksponent baze 2 u prikazu. */
    double signif;           /* Stvarni signifikand. */
} Number_Triplet;


/****************************************************************************/

    /* Pretvaranje prikaza broja u triplet. */

Number_Triplet Repr_to_Triplet(Number_Repr repr)
{
    Number_Triplet triplet;
    int bit, i, byte_index, byte_pos;
    unsigned char byte_mask;
    unsigned int charact = 0u;
    double signif = 0.0;

        /* Predznak u prikazu = vodeci bit. */
    byte_index = length - 1;
    byte_pos = 7;
    byte_mask = 0x80u;
    triplet.sign = repr.byte[byte_index] & byte_mask ? 1 : 0;

        /* Pomak bit-pozicije za jedno mjesto nanize. */
    --byte_pos;
    byte_mask = byte_mask >> 1;

        /* Slaganje karakteristike po bitovima - od najviseg,
           do najnizeg. Koristimo Hornerov algoritam u bazi 2. */
    for (i = exp_high_ind; i >= exp_low_ind; --i) {

        bit = repr.byte[byte_index] & byte_mask ? 1 : 0;

            /* Pomak karakteristike za bit navise i dodatak novog bita. */
        charact = (charact << 1) + bit;

            /* Pomak bit-pozicije za jedno mjesto nanize. */
        if (byte_pos > 0) {
            --byte_pos;
            byte_mask = byte_mask >> 1;
        }
        else {
            --byte_index;
            byte_pos = 7;
            byte_mask = 0x80u;
        }
    }

    triplet.charact = charact;

        /* Binarni eksponent = karakteritika - pomak (bias). */
    triplet.bin_exp = triplet.charact - bias;

        /* Slaganje signifikanda po bitovima - od najnizeg,
           do najviseg (koji moze biti skriven). */
    byte_index = 0;
    byte_pos = 0;
    byte_mask = 0x01u;
    for (i = 0; i < exp_low_ind; ++i) {

            /* Pomak signifikanda za bit nanize = podijeli ga s 2. */
        signif = 0.5 * signif;

            /* Provjera bita s indeksom i. Ako je jednak 1,
               onda dodaj 1 signifikandu. */
        if ((repr.byte[byte_index] & byte_mask) > 0)
            signif = signif + 1.0;

            /* Pomak bit-pozicije za jedno mjesto navise. */
        if (byte_pos < 7) {
            ++byte_pos;
            byte_mask = byte_mask << 1;
        }
        else {
            ++byte_index;
            byte_pos = 0;
            byte_mask = 0x01u;
        }
    }

        /* Ako je vodeci (cjelobrojni) bit signifikanda skriven,
           onda ga otkrivamo iz karakteritike/eksponenta. */
    if (hidden_bit) {
            /* Pomak signifikanda za bit nanize = podijeli ga s 2. */
        signif = 0.5 * signif;
            /* Provjera skrivenog bita iz karakteristike. Ako je
               karakteristika pozitivna, skriveni bit je jednak 1,
               onda dodaj 1 signifikandu. */
        if (triplet.charact > 0)
            signif = signif + 1.0;
    }

    triplet.signif = signif;

    return triplet;
}

/****************************************************************************/

    /* Pretvaranje zadanog tripleta u prikaz broja (uz kontrolu).
       Triplet mora imati zadanu karakteristiku, a binarni eksponent
       se ignorira (ne mora biti zadan). */

Number_Repr Triplet_to_Repr(Number_Triplet triplet)
{
    Number_Repr repr;
    int bit, i, byte_index, byte_pos;
    unsigned char byte_mask;
    unsigned int charact;

        /* Kontrola karakteristike. Provjerava se samo gornja granica,
           jer ulazna karakteristika ne moze biti negativna (unsigned). */

    if (triplet.charact > max_charact)
        triplet.charact = max_charact;

        /* Kontrola signifikanda. Za korektan ulaz, ocekujemo sljedece:

           - Ako je karakteristika pozitivna (i nije maksimalna), onda
             signifikand mora biti >= 1 i strogo manji od 2.
               Dakle, ocekujemo da signifikand (da na eventualni vodeci
             skriveni bit) zaista zadaje bitove u prikazu signifikanda.
             Onda ignoriramo eksponent ulaznog broja i doslovno kopiramo
             bitove iz prikaza signifikanda ulaznog broja.
               Za svaki slucaj, ovdje treba paziti na denormalizirane
             signifikande, i to samo kad cjelobrojni bit NIJE skriven,
             jer je u kontradikciji sa zadanom karakteristikom iz tripleta.

           - Ako je karakteristika = 0, onda je signifikand ili jednak nula,
             ili mora biti >= 2^(-prec_2) i strogo manji od 1, tako da
             dobijemo pozitivan broj.
               Ovdje ocekujemo da je ulazni signifikand = apsolutna vrijednost
             stvarnog signifikanda u prikazu broja. Umjesto denormalizacije
             ulaznog broja, bitove dobivamo tako da ulaznom broju dodamo 1
             i kopiramo bitove iza binarne tocke (bez vodece jedinice)!
               U slucaju da cjelobrojni bit NIJE skriven, treba taj bit
             korektno postaviti na nulu (kompenzacija za dodani 1).
        */

        /* Za svaki slucaj: ako je signif negativan, promijeni mu predznak. */
    if (triplet.signif < 0.0)
        triplet.signif = -triplet.signif;

        /* Iskopiraj zadani signifikand u broj. Kasnije cemo postaviti
           zadani predznak i karakteristiku. */
    repr.number = triplet.signif;

        /* Posebni slucaj kad je zadana karakteristika = 0, tj. triplet
           prikazuje denormalizirani broj ili 0. */
    if (triplet.charact == 0) {    /* Denormalizirani broj ili nula. */

            /* Za svaki slucaj: ako signif >= 1.0 (sto ne bi smio biti),
               uzmi samo racionalni dio signifikanda (tj. oduzmi floor). */
        if (triplet.signif >= 1.0)
            triplet.signif = triplet.signif - floor(triplet.signif);

            /* Sad dodamo 1, tako da je broj (trenutno) normaliziran,
               a zeljeni bitovi pravog signifikanda su pomaknuti tamo
               gdje treba. Kasnije cemo kompenzirati dodani 1. */
        repr.number = 1.0 + triplet.signif;
    }


        /* Slaganje karakteristike po bitovima - od najnizeg, do najviseg,
           tako da kopiramo bit-po-bit iz zadane karakteristike.
           Krecemo od najnizeg bita i postavljamo pripadne vrijednosti. */

    byte_index = exp_low_ind / 8;
    byte_pos = exp_low_ind % 8;
    byte_mask = 1u << ((unsigned) byte_pos);

        /* Kopija zadane karakteristike, kojoj cemo brisati bitove straga
           (tj. unistavamo najnizi bit). */
    charact = triplet.charact;

    for (i = exp_low_ind; i <= exp_high_ind; ++i) {

            /* Pripadni bit zadane karakteristike i brisanje tog bita
               iz "radne" karakteristike. */
        bit = charact % 2;
        charact = charact / 2;

            /* Postavi taj bit na pravo mjesto u prikazu. */
        if (bit == 0) {
                /* Ako je trenutni bit u prikazu jednak 1, postavi ga na 0. */
            if ((repr.byte[byte_index] & byte_mask) > 0)
                repr.byte[byte_index] = repr.byte[byte_index] - byte_mask;
        }
        else {
                /* Ako je trenutni bit u prikazu jednak 0, postavi ga na 1. */
            if ((repr.byte[byte_index] & byte_mask) == 0)
                repr.byte[byte_index] = repr.byte[byte_index] + byte_mask;
        }

            /* Pomak bit-pozicije za jedno mjesto navise. */
        if (byte_pos < 7) {
            ++byte_pos;
            byte_mask = byte_mask << 1;
        }
        else {
            ++byte_index;
            byte_pos = 0;
            byte_mask = 0x01u;
        }
    }

        /* Predznak u prikazu = vodeci bit. Uociti da su byte_index,
           byte_pos i byte_mask vec korektno postavljeni, jer se predznak
           nalazi jedan indeks vise od najviseg bita karakteristike.
             Ovdje radi i "skracena" varijanta koda (s operatorom | ),
           a nije potrebno eksplicitno testiranje bita (kao u komentaru). */

    if (triplet.sign != 0) {    /* Postavi znak na 1. */
        repr.byte[byte_index] = repr.byte[byte_index] | byte_mask;
/*
        if ((repr.byte[byte_index] & byte_mask) == 0)
            repr.byte[byte_index] = repr.byte[byte_index] + byte_mask;
*/
    }

        /* Sljedeci dio koda nije potreban za standardne IEEE tipove. */

        /* Na kraju, ako nema skrivenog bita, onda postavljamo vodeci
           (cjelobrojni) bit signifikanda, ovisno o karakteristici.
             Ako je zadana karakteristika = 0, taj bit postavljamo na 0.
           U protivnom, taj bit postavljamo na 1 (prema Intel standardu
           za extended double tip, da izbjegnemo nedozvoljeni prikaz).
             Taj ne-skriveni bit se nalazi na indeksu exp_low_ind - 1. */

    if (hidden_bit == 0) {
        i = exp_low_ind - 1;
        byte_index = i / 8;
        byte_pos = i % 8;
        byte_mask = 1u << ((unsigned) byte_pos);

        if (triplet.charact == 0) {    /* Denormalizirani broj ili nula. */

                /* Ako je skriveni bit u prikazu jednak 1, postavi ga na 0.
                   Ovo IMA smisla u nasoj implementaciji, jer smo zadanom
                   signifikandu (iz tripleta) dodali 1, da ga napravimo
                   normaliziranim. */
            if ((repr.byte[byte_index] & byte_mask) > 0)
                repr.byte[byte_index] = repr.byte[byte_index] - byte_mask;
        }
        else {    /* Normalizirani broj, Inf ili NaN. */

                /* Ako je skriveni bit u prikazu jednak 0, postavi ga na 1.
                   Ovo je samo IZUZETNI slucaj u nasoj implementaciji, ako
                   je zadani signifikand (iz tripleta) bio denormaliziran,
                   sto nema puno smisla, ali se moze dogoditi. */
            if ((repr.byte[byte_index] & byte_mask) == 0)
                repr.byte[byte_index] = repr.byte[byte_index] + byte_mask;
        }
    }

    return repr;
}

/****************************************************************************/

    /* Funkcija za ispis bitova u prikazu (sva tri dijela). */

/*
  Double:
    0 01111111111   1111 10000000 00000000 00000000 00000000 00000000 00000000
*/

void ispis_bitova(Number_Repr repr)
{
    int bit, i, byte_index, byte_pos;
    unsigned char byte_mask;

        /* Ispis bitova - od najviseg do najnizeg.
           Indeks najviseg bita u prikazu = ukupan broj bitova - 1. */
    byte_index = length - 1;
    byte_pos = 7;
    byte_mask = 0x80u;

        /* Vodeci bit = predznak. */
    bit = repr.byte[byte_index] & byte_mask ? 1 : 0;
    printf("    %1d ", bit);

        /* Pomak bit-pozicije za jedno mjesto nanize. */
    --byte_pos;
    byte_mask = byte_mask >> 1;

        /* Bitovi karakteristike. */
    for (i = exp_high_ind; i >= exp_low_ind; --i) {

        bit = repr.byte[byte_index] & byte_mask ? 1 : 0;
        printf("%1d", bit);

            /* Pomak bit-pozicije za jedno mjesto nanize. */
        if (byte_pos > 0) {
            --byte_pos;
            byte_mask = byte_mask >> 1;
        }
        else {
            --byte_index;
            byte_pos = 7;
            byte_mask = 0x80u;
        }
    }

    printf("   ");

        /* Bitovi signifikanda (bez skrivenog, ako ga ima). */
    for (i = exp_low_ind - 1; i >= 0; --i) {

        bit = repr.byte[byte_index] & byte_mask ? 1 : 0;
        printf("%1d", bit);

        if (i % 8 == 0 && i > 0) printf(" ");

            /* Pomak bit-pozicije za jedno mjesto nanize. */
        if (byte_pos > 0) {
            --byte_pos;
            byte_mask = byte_mask >> 1;
        }
        else {
            --byte_index;
            byte_pos = 7;
            byte_mask = 0x80u;
        }
    }

    printf("\n");

    return;
}

/****************************************************************************/

void ispis_prikaza(Number_Repr repr)
{
    Number_Triplet triplet;

        /* Pretvori zadani prikaz u triplet, za provjeru Inf/Nan. */
    triplet = Repr_to_Triplet(repr);


    printf("\n");

        /* Ispis kao broj. */
    printf("   Broj (%s) = ", real_type_name);

        /* Provjera za Inf/NaN. */
    if (triplet.charact == max_charact) {
        if (triplet.signif == 1.0) {    /* Inf. */
            if (triplet.sign == 0)
                printf("+Inf");
            else
                printf("-Inf");
        }
        else
            printf("NaN");
    }
    else {
            /* Korektni broj, ali moze biti triplet.charact = 0,
               tj. "gradual underflow". */
        printf("%*.*e",
               (int) (prec_10 + 7), (int) (prec_10 - 1), repr.number);
    }

    printf("\n");

        /* Ispis kao triplet. */
    printf("   Triplet = ( %1d, %5d [%4u], %*.*e )\n",
           triplet.sign, triplet.bin_exp, triplet.charact,
           (int) (prec_10 + 7), (int) (prec_10 - 1), triplet.signif);

        /* Ispis svih bitova u prikazu. */
    ispis_bitova(repr);

    return;
}

/****************************************************************************/

int main(void)
{
    Number_Repr repr;
    Number_Triplet triplet;

    triplet.bin_exp = 0;


        /* Primjer 1: Broj = 1. */
    printf("\n Broj = 1:\n");
    triplet.sign = 0;
    triplet.charact = bias;
    triplet.signif = 1.0;

    repr = Triplet_to_Repr(triplet);
    ispis_prikaza(repr);


        /* Primjer 2: Broj = 1 + 2u = 1 + prec = 1 + epsilon. */
    printf("\n Broj = 1 + 2u = 1 + prec = 1 + epsilon:\n");
    triplet.sign = 0;
    triplet.charact = bias;
    triplet.signif = 1.0 + pow(2, -((int)prec_2 - hidden_bit));

    repr = Triplet_to_Repr(triplet);
    ispis_prikaza(repr);


        /* Primjer 3: Broj = -1. */
    printf("\n Broj = -1:\n");
    triplet.sign = 1;
    triplet.charact = bias;
    triplet.signif = 1.0;

    repr = Triplet_to_Repr(triplet);
    ispis_prikaza(repr);


        /* Primjer 4: Najmanji pozitivni nenormalizirani broj. */
    printf("\n Najmanji pozitivni broj (nenormalizirani) = 2^(-1074):\n");
    triplet.sign = 0;
    triplet.charact = 0;
    triplet.signif = pow(2, -((int)prec_2 - hidden_bit));

    repr = Triplet_to_Repr(triplet);
    ispis_prikaza(repr);


        /* Primjer 5: Test za -Inf. */
    printf("\n Broj = -Inf:\n");
    triplet.sign = 1;
    triplet.charact = max_charact;
    triplet.signif = 1.0;

    repr = Triplet_to_Repr(triplet);
    ispis_prikaza(repr);


        /* Primjer 6: Test za NaN. */
    printf("\n Broj = NaN:\n");
    triplet.sign = 1;
    triplet.charact = max_charact;
    triplet.signif = 1.25;

    repr = Triplet_to_Repr(triplet);
    ispis_prikaza(repr);

    return 0;
}
