#include <stdio.h>
#include <math.h>
#include <float.h>

  /* Standardni parametri realnog tipa double (<float.h>, Lapack)
     i prikazi pripadnih brojeva. */

  /* Pretpostavka = "little endian" reprezentacija,
     tj. manje znacajni bitovi su na NIZIM adresama. */

  /* Napomena:
       Ovisno o opcijama kod prevodjenja, moze se dogoditi da funkcija printf
     NE zna korektno napisati broj. Primjer = 1.0e-310.
       Razlog: Opcija /O3 za optimizaciju ukljucuje i pretvaranje
     denormaliziranih brojeva u nulu ("flush denormal results to zero"). */


/****************************************************************************/


    /* String za ime realnog tipa. */
const char real_type_name[] = "double";


    /* Duljina tipa u byte-ovima. Broj bitova = length * 8. */
const unsigned int length = 8;

    /* Preciznost cijelog signifikanda (broj bitova = znamenki u bazi 2),
       ukljucivo i "skriveni" bit (hidden bit), ako postoji.
       Broj bitova u zavrsnom ("trailing", razlomljenom) dijelu
       signifikanda je t = prec_2 - 1, tj., najnizi bit signifikanda
       dolazi uz potenciju 2^(-t). */
const unsigned int prec_2 = 53;

    /* Potrebna preciznost u bazi 10 za korektan tzv.
       "write-read" ciklus, tako da dobijemo isti prikaz
       broja (u bazi 2), nakon ispisa i citanja (u bazi 10). */
const unsigned int prec_10 = 17;

    /* Logicki indikator = je li vodeci bit signifikanda "skriven"
       ili ne (tj. pamti se). Stvarni broj bitova za pamcenje
       signifikanda je q = prec_2 - hidden_bit. */
const int hidden_bit = 1;


    /* Broj bitova karakteristike, odnosno, binarnog eksponenta.
       Prema pretpostavci (little endian), ti bitovi se nalaze
       na najvisim adresama (odmah iza predznaka). */
const unsigned int exp_bits = 11;

    /* Maksimalna karakteristika = 2^(exp_bits) - 1. */
const unsigned int max_charact = 2047;

    /* Pomak za binarni eksponent = 2^(exp_bits - 1) - 1
       = max_charact / 2 (cjelobrojno dijeljenje). */
const int bias = 1023;


    /* Indeks bita predznaka (najvisi bit u cijelom prikazu)
       = length * 8 - 1, s tim da indekse brojimo od 0 (najnizi bit). */
const int sign_bit_ind = 63;

    /* Indeks vodeceg (najviseg) bita karakteristike (eksponenta)
       = sign_bit_ind - 1. */
const int exp_high_ind = 62;

    /* Indeks zadnjeg (najnizeg) bita karakteristike (eksponenta)
       = sign_bit_ind - exp_bits. */
const int exp_low_ind = 52;


/****************************************************************************/

    /* Tip za prikaz broja, tako da se lako dodje do bitova. */

typedef union {
    double number;            /* double = 64 bita. */
    unsigned char byte[8];    /* Polje od 8 byteova. */
} Number_Repr;

    /* Tip za vrijednosti pojedinih dijelova u prikazu broja,
       gledano kao triplet = (predznak, eksponent, signifikand).
       Uz stvarni binarni eksponent, pamtimo i karakteristiku. */

typedef struct {
    int sign;                /* Predznak = 0 ili 1. */
    unsigned int charact;    /* Karakteristika = pomaknuti eksponent. */
    int bin_exp;             /* Pravi eksponent baze 2 u prikazu. */
    double signif;           /* Stvarni signifikand. */
} Number_Triplet;


/****************************************************************************/

    /* Pretvaranje prikaza broja u triplet. */

Number_Triplet Repr_to_Triplet(Number_Repr repr)
{
    Number_Triplet triplet;
    int bit, i, byte_index, byte_pos;
    unsigned char byte_mask;
    unsigned int charact = 0u;
    double signif = 0.0;

        /* Predznak u prikazu = vodeci bit. */
    byte_index = length - 1;
    byte_pos = 7;
    byte_mask = 0x80u;
    triplet.sign = repr.byte[byte_index] & byte_mask ? 1 : 0;

        /* Pomak bit-pozicije za jedno mjesto nanize. */
    --byte_pos;
    byte_mask = byte_mask >> 1;

        /* Slaganje karakteristike po bitovima - od najviseg,
           do najnizeg. Koristimo Hornerov algoritam u bazi 2. */
    for (i = exp_high_ind; i >= exp_low_ind; --i) {

        bit = repr.byte[byte_index] & byte_mask ? 1 : 0;

            /* Pomak karakteristike za bit navise i dodatak novog bita. */
        charact = (charact << 1) + bit;

            /* Pomak bit-pozicije za jedno mjesto nanize. */
        if (byte_pos > 0) {
            --byte_pos;
            byte_mask = byte_mask >> 1;
        }
        else {
            --byte_index;
            byte_pos = 7;
            byte_mask = 0x80u;
        }
    }

    triplet.charact = charact;

        /* Binarni eksponent = karakteritika - pomak (bias). */
    triplet.bin_exp = triplet.charact - bias;

        /* Slaganje signifikanda po bitovima - od najnizeg,
           do najviseg (koji moze biti skriven). */
    byte_index = 0;
    byte_pos = 0;
    byte_mask = 0x01u;
    for (i = 0; i < exp_low_ind; ++i) {

            /* Pomak signifikanda za bit nanize = podijeli ga s 2. */
        signif = 0.5 * signif;

            /* Provjera bita s indeksom i. Ako je jednak 1,
               onda dodaj 1 signifikandu. */
        if ((repr.byte[byte_index] & byte_mask) > 0)
            signif = signif + 1.0;

            /* Pomak bit-pozicije za jedno mjesto navise. */
        if (byte_pos < 7) {
            ++byte_pos;
            byte_mask = byte_mask << 1;
        }
        else {
            ++byte_index;
            byte_pos = 0;
            byte_mask = 0x01u;
        }
    }

        /* Ako je vodeci (cjelobrojni) bit signifikanda skriven,
           onda ga otkrivamo iz karakteritike/eksponenta. */
    if (hidden_bit) {
            /* Pomak signifikanda za bit nanize = podijeli ga s 2. */
        signif = 0.5 * signif;
            /* Provjera skrivenog bita iz karakteristike. Ako je
               karakteristika pozitivna, skriveni bit je jednak 1,
               onda dodaj 1 signifikandu. */
        if (triplet.charact > 0)
            signif = signif + 1.0;
    }

    triplet.signif = signif;

    return triplet;
}

/****************************************************************************/

    /* Pretvaranje zadanog tripleta u prikaz broja (uz kontrolu).
       Triplet mora imati zadanu karakteristiku, a binarni eksponent
       se ignorira (ne mora biti zadan). */

Number_Repr Triplet_to_Repr(Number_Triplet triplet)
{
    Number_Repr repr;
    int bit, i, byte_index, byte_pos;
    unsigned char byte_mask;
    unsigned int charact;

        /* Kontrola karakteristike. Provjerava se samo gornja granica,
           jer ulazna karakteristika ne moze biti negativna (unsigned). */

    if (triplet.charact > max_charact)
        triplet.charact = max_charact;

        /* Kontrola signifikanda. Za korektan ulaz, ocekujemo sljedece:

           - Ako je karakteristika pozitivna (i nije maksimalna), onda
             signifikand mora biti >= 1 i strogo manji od 2.
               Dakle, ocekujemo da signifikand (da na eventualni vodeci
             skriveni bit) zaista zadaje bitove u prikazu signifikanda.
             Onda ignoriramo eksponent ulaznog broja i doslovno kopiramo
             bitove iz prikaza signifikanda ulaznog broja.
               Za svaki slucaj, ovdje treba paziti na denormalizirane
             signifikande, i to samo kad cjelobrojni bit NIJE skriven,
             jer je u kontradikciji sa zadanom karakteristikom iz tripleta.

           - Ako je karakteristika = 0, onda je signifikand ili jednak nula,
             ili mora biti >= 2^(-prec_2) i strogo manji od 1, tako da
             dobijemo pozitivan broj.
               Ovdje ocekujemo da je ulazni signifikand = apsolutna vrijednost
             stvarnog signifikanda u prikazu broja. Umjesto denormalizacije
             ulaznog broja, bitove dobivamo tako da ulaznom broju dodamo 1
             i kopiramo bitove iza binarne tocke (bez vodece jedinice)!
               U slucaju da cjelobrojni bit NIJE skriven, treba taj bit
             korektno postaviti na nulu (kompenzacija za dodani 1).
        */

        /* Za svaki slucaj: ako je signif negativan, promijeni mu predznak. */
    if (triplet.signif < 0.0)
        triplet.signif = -triplet.signif;

        /* Iskopiraj zadani signifikand u broj. Kasnije cemo postaviti
           zadani predznak i karakteristiku. */
    repr.number = triplet.signif;

        /* Posebni slucaj kad je zadana karakteristika = 0, tj. triplet
           prikazuje denormalizirani broj ili 0. */
    if (triplet.charact == 0) {    /* Denormalizirani broj ili nula. */

            /* Za svaki slucaj: ako signif >= 1.0 (sto ne bi smio biti),
               uzmi samo racionalni dio signifikanda (tj. oduzmi floor). */
        if (triplet.signif >= 1.0)
            triplet.signif = triplet.signif - floor(triplet.signif);

            /* Sad dodamo 1, tako da je broj (trenutno) normaliziran,
               a zeljeni bitovi pravog signifikanda su pomaknuti tamo
               gdje treba. Kasnije cemo kompenzirati dodani 1. */
        repr.number = 1.0 + triplet.signif;
    }


        /* Slaganje karakteristike po bitovima - od najnizeg, do najviseg,
           tako da kopiramo bit-po-bit iz zadane karakteristike.
           Krecemo od najnizeg bita i postavljamo pripadne vrijednosti. */

    byte_index = exp_low_ind / 8;
    byte_pos = exp_low_ind % 8;
    byte_mask = 1u << ((unsigned) byte_pos);

        /* Kopija zadane karakteristike, kojoj cemo brisati bitove straga
           (tj. unistavamo najnizi bit). */
    charact = triplet.charact;

    for (i = exp_low_ind; i <= exp_high_ind; ++i) {

            /* Pripadni bit zadane karakteristike i brisanje tog bita
               iz "radne" karakteristike. */
        bit = charact % 2;
        charact = charact / 2;

            /* Postavi taj bit na pravo mjesto u prikazu. */
        if (bit == 0) {
                /* Ako je trenutni bit u prikazu jednak 1, postavi ga na 0. */
            if ((repr.byte[byte_index] & byte_mask) > 0)
                repr.byte[byte_index] = repr.byte[byte_index] - byte_mask;
        }
        else {
                /* Ako je trenutni bit u prikazu jednak 0, postavi ga na 1. */
            if ((repr.byte[byte_index] & byte_mask) == 0)
                repr.byte[byte_index] = repr.byte[byte_index] + byte_mask;
        }

            /* Pomak bit-pozicije za jedno mjesto navise. */
        if (byte_pos < 7) {
            ++byte_pos;
            byte_mask = byte_mask << 1;
        }
        else {
            ++byte_index;
            byte_pos = 0;
            byte_mask = 0x01u;
        }
    }

        /* Predznak u prikazu = vodeci bit. Uociti da su byte_index,
           byte_pos i byte_mask vec korektno postavljeni, jer se predznak
           nalazi jedan indeks vise od najviseg bita karakteristike.
             Ovdje radi i "skracena" varijanta koda (s operatorom | ),
           a nije potrebno eksplicitno testiranje bita (kao u komentaru). */

    if (triplet.sign != 0) {    /* Postavi znak na 1. */
        repr.byte[byte_index] = repr.byte[byte_index] | byte_mask;
/*
        if ((repr.byte[byte_index] & byte_mask) == 0)
            repr.byte[byte_index] = repr.byte[byte_index] + byte_mask;
*/
    }

        /* Sljedeci dio koda nije potreban za standardne IEEE tipove. */

        /* Na kraju, ako nema skrivenog bita, onda postavljamo vodeci
           (cjelobrojni) bit signifikanda, ovisno o karakteristici.
             Ako je zadana karakteristika = 0, taj bit postavljamo na 0.
           U protivnom, taj bit postavljamo na 1 (prema Intel standardu
           za extended double tip, da izbjegnemo nedozvoljeni prikaz).
             Taj ne-skriveni bit se nalazi na indeksu exp_low_ind - 1. */

    if (hidden_bit == 0) {
        i = exp_low_ind - 1;
        byte_index = i / 8;
        byte_pos = i % 8;
        byte_mask = 1u << ((unsigned) byte_pos);

        if (triplet.charact == 0) {    /* Denormalizirani broj ili nula. */

                /* Ako je skriveni bit u prikazu jednak 1, postavi ga na 0.
                   Ovo IMA smisla u nasoj implementaciji, jer smo zadanom
                   signifikandu (iz tripleta) dodali 1, da ga napravimo
                   normaliziranim. */
            if ((repr.byte[byte_index] & byte_mask) > 0)
                repr.byte[byte_index] = repr.byte[byte_index] - byte_mask;
        }
        else {    /* Normalizirani broj, Inf ili NaN. */

                /* Ako je skriveni bit u prikazu jednak 0, postavi ga na 1.
                   Ovo je samo IZUZETNI slucaj u nasoj implementaciji, ako
                   je zadani signifikand (iz tripleta) bio denormaliziran,
                   sto nema puno smisla, ali se moze dogoditi. */
            if ((repr.byte[byte_index] & byte_mask) == 0)
                repr.byte[byte_index] = repr.byte[byte_index] + byte_mask;
        }
    }

    return repr;
}

/****************************************************************************/

    /* Funkcija za ispis bitova u prikazu (sva tri dijela). */

/*
  Double:
    0 01111111111   1111 10000000 00000000 00000000 00000000 00000000 00000000
*/

void ispis_bitova(Number_Repr repr)
{
    int bit, i, byte_index, byte_pos;
    unsigned char byte_mask;

        /* Ispis bitova - od najviseg do najnizeg.
           Indeks najviseg bita u prikazu = ukupan broj bitova - 1. */
    byte_index = length - 1;
    byte_pos = 7;
    byte_mask = 0x80u;

        /* Vodeci bit = predznak. */
    bit = repr.byte[byte_index] & byte_mask ? 1 : 0;
    printf("    %1d ", bit);

        /* Pomak bit-pozicije za jedno mjesto nanize. */
    --byte_pos;
    byte_mask = byte_mask >> 1;

        /* Bitovi karakteristike. */
    for (i = exp_high_ind; i >= exp_low_ind; --i) {

        bit = repr.byte[byte_index] & byte_mask ? 1 : 0;
        printf("%1d", bit);

            /* Pomak bit-pozicije za jedno mjesto nanize. */
        if (byte_pos > 0) {
            --byte_pos;
            byte_mask = byte_mask >> 1;
        }
        else {
            --byte_index;
            byte_pos = 7;
            byte_mask = 0x80u;
        }
    }

    printf("   ");

        /* Bitovi signifikanda (bez skrivenog, ako ga ima). */
    for (i = exp_low_ind - 1; i >= 0; --i) {

        bit = repr.byte[byte_index] & byte_mask ? 1 : 0;
        printf("%1d", bit);

        if (i % 8 == 0 && i > 0) printf(" ");

            /* Pomak bit-pozicije za jedno mjesto nanize. */
        if (byte_pos > 0) {
            --byte_pos;
            byte_mask = byte_mask >> 1;
        }
        else {
            --byte_index;
            byte_pos = 7;
            byte_mask = 0x80u;
        }
    }

    printf("\n");

    return;
}

/****************************************************************************/

void ispis_prikaza(Number_Repr repr)
{
    Number_Triplet triplet;

        /* Pretvori zadani prikaz u triplet, za provjeru Inf/Nan. */
    triplet = Repr_to_Triplet(repr);


    printf("\n");

        /* Ispis kao broj. */
    printf("   Broj (%s) = ", real_type_name);

        /* Provjera za Inf/NaN. */
    if (triplet.charact == max_charact) {
        if (triplet.signif == 1.0) {    /* Inf. */
            if (triplet.sign == 0)
                printf("+Inf");
            else
                printf("-Inf");
        }
        else
            printf("NaN");
    }
    else {
            /* Korektni broj, ali moze biti triplet.charact = 0,
               tj. "gradual underflow". */
        printf("%*.*e",
               (int) (prec_10 + 7), (int) (prec_10 - 1), repr.number);
    }

    printf("\n");

        /* Ispis kao triplet. */
    printf("   Triplet = ( %1d, %5d [%4u], %*.*e )\n",
           triplet.sign, triplet.bin_exp, triplet.charact,
           (int) (prec_10 + 7), (int) (prec_10 - 1), triplet.signif);

        /* Ispis svih bitova u prikazu. */
    ispis_bitova(repr);

    return;
}

/****************************************************************************/

int main(void)
{
    int I_fmt = 5;       /* Za e_min, s predznakom. */
    int F_fmt = I_fmt + 2;
    int E_fmt_1 = 24;    /* = (int) (prec_10 + 7); */
    int E_fmt_2 = 16;    /* = (int) (prec_10 - 1); */

        /* Osnovni cjelobrojni parametri prikaza. */
    int base, p, t, q, w, e_true_min, e_true_max, e_min, e_max, rnd;

        /* Osnovni parametri kao varijable tipa Number_Repr. */
    Number_Repr One, Prec, Eps, Rmin, RTmin, Rmax, Smin, Smax, Tol;

        /* Osnovni parametri kao realne vrijednosti (Lapack stil). */
    double Mach_Prec, Mach_Eps, Mach_Rmin, Mach_RTmin, Mach_Rmax;
    double Mach_Smin, Mach_Smax, Mach_Tol;

        /* Cjelobrojni parametri kao realne vrijednosti (Lapack stil). */
    double Mach_Base, Mach_P, Mach_T, Mach_Q, Mach_ETmin, Mach_ETmax;
    double Mach_Emin, Mach_Emax, Mach_Rnd;

        /* Pomocne varijable. */
    Number_Triplet triplet;
    Number_Repr One_plus_prec, pom_repr;
    int i;


        /* Binarni eksponent se ignorira u tripletu,
           gleda se samo karakteristika. */
    triplet.bin_exp = 0;


      /***** Cjelobrojni parametri. *****/

        /* Baza prikaza za tip (radix). */
    base = 2;

        /* Broj znacajnih znamenki signifikanda (p = preciznost). */
    p = (int) prec_2;

        /* Broj znamenki u razlomljenom dijelu signifikanda.
           Najniza znamenka signifikanda ide uz potenciju base^(-t). */
    t = p - 1;

        /* Stvarni broj znamenki signifikanda koji se pamti u prikazu. */
    q = p - hidden_bit;

        /* Broj znamenki karakteristike/eksponenta. */
    w = (int) exp_bits;

        /* Stvarni najmanji i najveci eksponent baze u prikazu. */
    e_true_min = 1 - bias;
    e_true_max = (int) max_charact - bias - 1;

        /* Najmanji i najveci eksponent baze (Lapack stil), gledano
           na zapis broja u obliku  0.mantisa * baza^(eksponent).
           Ovi brojevi su za  1 veci od pripadnih stvarnih. */
    e_min = 1 + e_true_min;
    e_max = 1 + e_true_max;

        /* Indikator zaokruzivanja = je li zaokruzivanje korektno,
           tj. na najblizi prikazavi broj, ili ne (onda moze biti
           prema nuli (tj. prema dolje, po apsolutnoj vrijednosti),
           ili afino - prema gore (Inf), odnosno, prema dolje (-Inf). */

    rnd = 1;    /* Korektno zaokruzivanje na najblizi. */


        /* Ispis cjelobrojnih parametara. */
    printf("\n");
    printf(" Cjelobrojni parametri prikaza tipa %s:\n", real_type_name);
    printf("\n");
    printf("   Baza prikaza = %*d  (_DBL_RADIX = %d)\n",
           I_fmt, base, _DBL_RADIX);
    printf("   Preciznost p = %*d  (DBL_MANT_DIG = %d)\n",
           I_fmt, p, DBL_MANT_DIG);
    printf("   Duljina razlomljenog dijela t = %*d\n",
           I_fmt, t);
    printf("   Duljina signifikanda  q = %*d\n",
           I_fmt, q);
    printf("   Duljina karakteritike w = %*d\n",
           I_fmt, w);
    printf("   Najmanji eksponent baze e_true_min = %*d\n",
           I_fmt, e_true_min);
    printf("   Najveci eksponent baze  e_true_max = %*d\n",
           I_fmt, e_true_max);
    printf("   Najmanji eksponent (Lapack)  e_min = %*d  (DBL_MIN_EXP = %d)\n",
           I_fmt, e_min, DBL_MIN_EXP);
    printf("   Najveci eksponent (Lapack)   e_max = %*d  (DBL_MAX_EXP = %d)\n",
           I_fmt, e_max, DBL_MAX_EXP);
    printf("   Indikator zaokruzivanja rnd = %*d\n",
           I_fmt, rnd);


      /***** Realni parametri. *****/

    printf("\n\n");
    printf(" Realni parametri prikaza tipa %s:\n", real_type_name);
    printf("\n");


        /* Osnovna velicina: Broj = 1. */

    printf("\n");
    printf(" Broj = 1:\n");

    triplet.sign = 0;
    triplet.charact = bias;
    triplet.signif = 1.0;

    One = Triplet_to_Repr(triplet);
    ispis_prikaza(One);


        /* Najmanji egzaktno prikazivi broj koji je strogo veci od 1.
           Dobivamo ga tako da zadnji bit signifikanda postavimo na 1.
           Taj broj pisemo u obliku 1 + prec (ili 1 + epsilon), gdje
           je prec (epsilon) egzaktno prikaziv.
           Standardno, u bazi 2, je prec = epsilon = 2^(-p + 1). */

    printf("\n");
    printf(" Najmanji prikazivi broj 1 + prec (1 + epsilon) > 1:\n");

    One_plus_prec.number = One.number;
    One_plus_prec.byte[0] = 0x01;

    ispis_prikaza(One_plus_prec);


        /* Najmanji prikazivi broj prec, takav da je broj 1 + prec egzaktno
           prikaziv i da je 1 + prec > 1. Ovdje NEMA zbrajanja u 1 + prec,
           vec se gleda samo prikaz broja. Katkad se prec zove epsilon.
           Dobivamo ga oduzimanjem (1 + prec) - 1 (koje je egzaktno). */

    printf("\n");
    printf(" Pripadni broj prec = epsilon  (DBL_EPSILON = %*.*e):\n",
           E_fmt_1, E_fmt_2, DBL_EPSILON);

    Prec.number = One_plus_prec.number - One.number;

    ispis_prikaza(Prec);


        /* Provjera zaokruzivanja i jedinicna greska zaokruzivanja. */

    printf("\n\n");
    printf(" Provjera zaokruzivanja:\n");

        /* Prvo izracunamo 1 + prec/2.
           Ako je to jednako 1, zaokruzivanje NIJE prema gore (-> Inf),
           jer bi rezultat bio l + prec > 1.
             Kod korektnog zaokruzivanja, rezultat za 1 + prec/2 je 1,
           zbog zaokruzivanja na "parni" zadnji bit, kad su dva prikaziva
           susjeda (ovdje 1 i 1 + prec) jednako udaljena od rezultata.
           Dakle, treba provjeriti preostale 3 mogucnosti.

           Onda izracunamo 1 + 0.75 * prec (moze i 1 + 0.55 * prec).
           Ako to daje 1 + prec, onda je sigurno korektno zaokruzivanje.
             Naime, za preostale dvije mogucnosti (-> 0, -> -Inf), rezultat
           mora biti 1.

           Iza toga postavimo rnd i postavimo u na prec/2 ili na prec. */

    printf("\n");
    printf(" Pomocni broj_1 = 1 + prec / 2:\n");

    pom_repr.number = One.number + Prec.number / 2;
    ispis_prikaza(pom_repr);

    if (pom_repr.number > One.number) rnd = 0;    /* Prema +Inf. */


    printf("\n");
    printf(" Pomocni broj_2 = 1 + 0.75 * prec:\n");

    pom_repr.number = One.number + 0.75 * Prec.number;
    ispis_prikaza(pom_repr);

    if (pom_repr.number == One.number) rnd = 0;    /* Prema 0 ili -Inf. */


    printf("\n");
    printf(" Provjera zaokruzivanja: rnd = %*d\n",
           I_fmt, rnd);


        /* Jedinicna greska zaokruzivanja u, koja se jos zove i eps.
           U bazi 2, kod korektnog zaokruzivanja na najblizi (rnd = 1) je
           u = eps = prec/2, tj. u = 2^{-p), a kod ostalih zaokruzivanja
           je u = eps = prec = 2^{-p + 1). */

    printf("\n\n");
    printf(" Jedinicna greska zaokruzivanja u = eps:\n");

    if (rnd)
        Eps.number = Prec.number / 2;
    else
        Eps.number = Prec.number;

    ispis_prikaza(Eps);


        /* Najmanji pozitivni normalizirani broj. */

    printf("\n");
    printf(" Donja granica tocnih prikaza (\"underflow threshold\")\n");
    printf(" Najmanji normalizirani broj Rmin > 0  (DBL_MIN = %*.*e):\n",
           E_fmt_1, E_fmt_2, DBL_MIN);

    triplet.sign = 0;
    triplet.charact = 1u;
    triplet.signif = 1.0;

    Rmin = Triplet_to_Repr(triplet);
    ispis_prikaza(Rmin);


        /* Najmanji pozitivni broj (nenormaliziran). */

    printf("\n");
    printf(" Najmanji broj RTmin > 0  (DBL_TRUE_MIN = %*.*e):\n",
           E_fmt_1, E_fmt_2, DBL_TRUE_MIN);

    RTmin.number = 0.0;
    RTmin.byte[0] = 0x01;

    ispis_prikaza(RTmin);


        /* Najveci pozitivni broj (strogo manji od +Inf). */

    printf("\n");
    printf(" Gornja granica tocnih prikaza (\"overflow threshold\")\n");
    printf(" Najveci broj Rmax  (DBL_MAX = %*.*e):\n",
           E_fmt_1, E_fmt_2, DBL_MAX);

    triplet.sign = 0;
    triplet.charact = max_charact - 1u;
    triplet.signif = 2.0 - Prec.number;

    Rmax = Triplet_to_Repr(triplet);
    ispis_prikaza(Rmax);


        /* Reciprocna vrijednost najmanjeg pozitivnog normaliziranog broja,
           Smax = 1 / Rmin. Po IEEE, mora biti prikaziva. */

    printf("\n");
    printf(" Reciprocna vrijednost donje granice, Smax = 1 / Rmin:\n");

    Smax.number = 1.0 / Rmin.number;

    ispis_prikaza(Smax);


        /* Reciprocna vrijednost najveceg pozitivnog broja,
           Smin = 1 / Rmax. Po IEEE, zalazi u undeflow. */

    printf("\n");
    printf(" Reciprocna vrijednost gornje granice, Smin = 1 / Rmax:\n");

    Smin.number = 1.0 / Rmax.number;

    ispis_prikaza(Smin);


        /* Lapackov parametar Tol = underflow_threshold (Rmin) / Prec. */

    printf("\n");
    printf(" Lapackov numericki parametar Tol = Rmin / Prec:\n");

    Tol.number = Rmin.number / Prec.number;

    ispis_prikaza(Tol);


/****************************************************************************/

        /* Machine precision data (Lapack style). */

    Mach_Base  = base;
    Mach_P     = p;
    Mach_T     = t;
    Mach_Q     = q;
    Mach_ETmin = e_true_min;
    Mach_ETmax = e_true_max;
    Mach_Emin  = e_min;
    Mach_Emax  = e_max;
    Mach_Rnd   = rnd;

    Mach_Eps   = Eps.number;
    Mach_Prec  = Prec.number;
    Mach_Rmin  = Rmin.number;
    Mach_RTmin = RTmin.number;
    Mach_Rmax  = Rmax.number;
    Mach_Smin  = Smin.number;
    Mach_Smax  = Smax.number;
    Mach_Tol   = Tol.number;

    printf("\n\n ");
    for (i = 1; i <= 78; ++i) printf("=");
    printf("\n\n");

    printf("        Mach_Base  = %*.1f;\n", F_fmt, Mach_Base);
    printf("        Mach_P     = %*.1f;\n", F_fmt, Mach_P);
    printf("        Mach_T     = %*.1f;\n", F_fmt, Mach_T);
    printf("        Mach_Q     = %*.1f;\n", F_fmt, Mach_Q);
    printf("        Mach_ETmin = %*.1f;\n", F_fmt, Mach_ETmin);
    printf("        Mach_ETmax = %*.1f;\n", F_fmt, Mach_ETmax);
    printf("        Mach_Emin  = %*.1f;\n", F_fmt, Mach_Emin);
    printf("        Mach_Emax  = %*.1f;\n", F_fmt, Mach_Emax);
    printf("        Mach_Rnd   = %*.1f;\n", F_fmt, Mach_Rnd);

    printf("        Mach_Eps   = %*.*e;\n", E_fmt_1, E_fmt_2, Mach_Eps);
    printf("        Mach_Prec  = %*.*e;\n", E_fmt_1, E_fmt_2, Mach_Prec);
    printf("        Mach_Rmin  = %*.*e;\n", E_fmt_1, E_fmt_2, Mach_Rmin);
    printf("        Mach_RTmin = %*.*e;\n", E_fmt_1, E_fmt_2, Mach_RTmin);
    printf("        Mach_Rmax  = %*.*e;\n", E_fmt_1, E_fmt_2, Mach_Rmax);
    printf("        Mach_Smin  = %*.*e;\n", E_fmt_1, E_fmt_2, Mach_Smin);
    printf("        Mach_Smax  = %*.*e;\n", E_fmt_1, E_fmt_2, Mach_Smax);
    printf("        Mach_Tol   = %*.*e;\n", E_fmt_1, E_fmt_2, Mach_Tol);

    return 0;
}
